"""
Clarisse

A light-weighted GUI framework for Python programs.

by 1MLightyears@gmail.com

on 20201208
"""
from __future__ import absolute_import

import os
import sys
from functools import wraps

from GUI import ClrsGUI

#from .analyze import AnalyzeFunc


__all__ = ["Clarisse", "version"]

version="v0.0.2"

def Clarisse(*args, **kwargs):
    def middle(func):
        if isinstance(func,type(lambda :0)):
            @wraps(func)
            def gui_func(*default_args, **default_kwargs):
                nonlocal func,args,kwargs
                w=ClrsGUI()
                w.Parse(func,default_args,default_kwargs,*args,**kwargs)
                w.main_window.show()
                w.app.exec_()
                return w.pages[0].run_thread.ret
            return gui_func
    return middle





"""
Clrsarisse

types_supported module.
Defines all types that are able to be displayed by PyQt widgets.

by 1MLightyears@gmail.com

on 20201208
"""

from PySide2.QtWidgets import QLabel, QSpinBox, QLineEdit, QDoubleSpinBox, QTextEdit, QTableWidget, QCheckBox, QTableWidgetItem
from PySide2.QtCore import Qt
import typing

import log

__all__ = ["ClrsUnknown","ClrsInt", "ClrsFloat","ClrsString","ClrsList","ClrsBool","ClrsDict"]

class ClrsUnknown(QLineEdit):
    __name__ = "ClrsUnknown"

    def __init__(self, name:str="",target_kwargs:dict={},*args,**kwargs):
        super(ClrsUnknown, self).__init__(*args, **kwargs)
        if name!="":
            self.setObjectName(name)
        self.target_kwargs = target_kwargs

    def setDefault(self, default=None):
        if default!=None:
            self.setPlaceholderText(str(default))

    def getValue(self):
        return self.text() or self.placeholderText()

class ClrsString(QLineEdit):
    __name__ = "ClrsString"

    def __init__(self, name:str="",target_kwargs:dict={},*args,**kwargs):
        super(ClrsString, self).__init__(*args, **kwargs)
        if name!="":
            self.setObjectName(name)
        self.target_kwargs = target_kwargs

    def setDefault(self, default=None):
        if default!=None:
            self.setPlaceholderText(default)

    def getValue(self):
        return self.text() or self.placeholderText()

class ClrsInt(QSpinBox):
    __name__ = "ClrsInt"

    def __init__(self, name:str="",target_kwargs:dict={},*args,**kwargs):
        super(ClrsInt, self).__init__(*args, **kwargs)
        if name!="":
            self.setObjectName(name)
        self.target_kwargs = target_kwargs

    def setDefault(self, default=None):
        if default != None:
            self.setValue(default)

    def getValue(self):
        return self.value()

class ClrsFloat(QDoubleSpinBox):
    __name__ = "ClrsFloat"

    def __init__(self, name:str="",target_kwargs:dict={},*args,**kwargs):
        super(ClrsFloat, self).__init__(*args, **kwargs)
        if name!="":
            self.setObjectName(name)
        self.target_kwargs = target_kwargs

    def setDefault(self, default=None):
        if default!=None:
            self.setValue(default)

    def getValue(self):
        return self.value()

class ClrsList(QTextEdit):
    __name__ = "ClrsList"

    def __init__(self,
            name: str = "",
            target_kwargs: dict = {},
            ElementType: type = None,
            *args,**kwargs):
        super(ClrsList, self).__init__(*args, **kwargs)
        if name!="":
            self.setObjectName(name)
        self.target_kwargs = target_kwargs
        self.ElementType = ElementType

    def setDefault(self, default=None):
        if isinstance(default,list):
            self.setPlaceholderText("\n".join([str(i) for i in default]))

    def getValue(self):
        t = self.placeholderText() if self.toPlainText() == "" else self.toPlainText()
        if (isinstance(self.ElementType,tuple))and(len(self.ElementType)>0):
            return [self.ElementType(i) for i in t.split("\n")]
        return [i for i in t.split("\n")]

class ClrsBool(QCheckBox):
    __name__ = "ClrsBool"

    def __init__(self, name:str="",target_kwargs:dict={},*args,**kwargs):
        super(ClrsBool, self).__init__(*args, **kwargs)
        if name!="":
            self.setObjectName(name)
        self.target_kwargs = target_kwargs

    def setDefault(self, default=None):
        if default!=None:
            self.setChecked(default)

    def getValue(self):
        return self.isChecked()

class ClrsDict(QTableWidget):
    __name__ = "ClrsDict"

    def __init__(self,
            name: str = "",
            target_kwargs: dict = {},
            ElementType = None,
            *args, **kwargs
        ):
        super(ClrsDict, self).__init__(*args, **kwargs)
        if name!="":
            self.setObjectName(name)
        self.target_kwargs = target_kwargs
        self.setDefaultDropAction(Qt.TargetMoveAction)
        self.ElementType = ElementType

    def setDefault(self, default=None):
        if default != None:
            # create a 2 column table
            keys = list(default.keys())
            keys_count = len(keys)
            self.setRowCount(keys_count)
            self.setColumnCount(2)
            for i in range(keys_count):
                k, v = QTableWidgetItem(), QTableWidgetItem()
                k.setText(str(keys[i])); v.setText(str(default[keys[i]]))
                self.setItem(i, 0, k)
                self.setItem(i, 1, v)
        # TODO:add approaches to add/delete items

    def getValue(self):
        ret = {}
        for i in range(self.rowCount()):
            k, v = self.item(i, 0).text(), self.item(i, 1).text()
            if isinstance(self.ElementType,tuple):
                if len(self.ElementType)>0:
                    k=self.ElementType[0](k)
                if len(self.ElementType)>1:
                    v=self.ElementType[1](v)
            ret.update({k:v})
        return ret


"""
Clarisse

Window drawing module.
analyze the interface of a function/class.

by 1MLightyears@gmail.com

on 20201211
"""
from analyze import AnalyzeFunc
from page import Page

import sys
import log

from PySide2.QtWidgets import QMainWindow, QApplication, QLabel,  QTabWidget,  QWidget
from PySide2.QtCore import  Signal, QCoreApplication, QMetaObject, QSize
from PySide2.QtGui import  QColor

__all__=["ClrsGUI"]

class Ui_MainWindow(QMainWindow, object):
    OnResize=Signal(QSize)
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(320,480)

        self.retranslateUi(MainWindow)
        QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))

    def resizeEvent(self,event):
        QMainWindow.resizeEvent(self, event)
        self.OnResize.emit(self.size())


class ClrsGUI():
    def __init__(self):
        """
        Create and set the Main_Window.
        """
        log.info("ClrsGUI init")
        self.app = QApplication(sys.argv)
        self.main_window = Ui_MainWindow()
        self.main_window.setupUi(self.main_window)
        log.info("main_window create size={size}".format(size=self.main_window.geometry()))

        self.central_widget = QWidget(self.main_window)
        self.main_window.setCentralWidget(self.central_widget)
        self.main_window.OnResize.connect(self.ReDrawAllPages)
        self.main_window.setMinimumWidth(320)

    def Parse(self,
            backend,
            default_args: tuple = (),
            default_kwargs: dict = {},
            args_desc: list = [],
            args_kwdesc: dict = {},
            *args,**kwargs):
        """
        Parse the decorated.
        """
        self.pages = []
        self.tabs = None

        # if backend is a function
        if isinstance(backend, type(lambda: 0)):
            # then args and kwargs store the default of all variables
            self.pages.append(Page(backend,*args,**kwargs))
            self.pages[-1].setParent(self.central_widget)
            self.DrawFuncInPage(self.pages[0], backend,default_args,default_kwargs,args_desc,args_kwdesc)
            self.main_window.resize(
                max([i.description.width() + i.description.x() + i.margin for i in self.pages]),
                self.main_window.height()
            )  # resize main_window's width to show all the description
            self.main_window.setWindowTitle("{0} - Clarisse GUI".format(backend.__name__))
        elif isinstance(a, object):
            self.tabs = QTabWidget(self.central_widget)
            # TODO:fill the codes; support classes
            # if backend is a class, decorate its __init__, make all its callable public
            # a page and organize them with self.tabs

    def DrawFuncInPage(self,
            page,
            func,
            default_args: tuple = (),
            default_kwargs: dict = {},
            args_desc: list = [],
            args_kwdesc: dict = {}
        ):
        """
        Draw parameters of function(func) in the main_window,
        set them with appropriate initial values.
        """
        # inspect func
        page.widget_list = AnalyzeFunc(func)
        log.info("page name={0},widget_list={1}".format(page.objectName(),[i.__name__ for i in page.widget_list]))

        # add widgets to canvas
        if page.widget_list != []:
            top=0
            for i in range(len(page.widget_list)):
                row_desc = page.widget_list[i].objectName()
                if len(args_desc) > 0:
                    row_desc = args_desc[0]
                    args_desc = args_desc[1:]
                if page.widget_list[i].objectName() in args_kwdesc:
                    row_desc = args_kwdesc[page.widget_list[i].objectName()]
                page.form_layout.addRow(QLabel(row_desc), page.widget_list[i])

                # apply arguments given to the original func as default args
                if len(default_args)>0:
                    page.widget_list[i].setDefault(default_args[0])
                    default_args = default_args[1:]
                if page.widget_list[i].objectName() in default_kwargs:
                    page.widget_list[i].setDefault(default_args[page.widget_list[i].objectName()])

                log.info("add widget name={name}, geometry={geometry}, desc={desc}".format(
                    name=page.widget_list[i].objectName(),
                    geometry=page.widget_list[i].geometry(),
                    desc=row_desc))

        # canvas resize
        page.canvas.resize(
            page.width(),
            page.canvas.y()+page.widget_list[-1].y()+page.widget_list[-1].height()+page.margin
        )
        log.info("canvas:{0}".format(page.canvas.geometry()))
        log.info("widget_scroll geometry={0}".format(page.widget_scroll.geometry()))
        log.info("page geometry={0}".format(page.geometry()))

        # modify Z-order
        page.run_button.raise_()
        page.description.raise_()

    def ReDrawAllPages(self, new_size: QSize = QSize(640,480)):
        """
        redraw each page to fit the form's size.
        """
        # tabs should adjust size
        if self.tabs != None:
            self.tabs.resize(new_size)
        for page in self.pages:
            page.Current_Layout(new_size)

        log.info("resize all pages to {0}".format(self.central_widget.geometry()))




"""
Clarisse

page module.
Define class Page, the canvas of type in types_supported.py.

by 1MLightyears@gmail.com

on 20201211
"""

from PySide2.QtWidgets import QPushButton,QScrollArea,QLineEdit,QLabel,QWidget,QFormLayout
from PySide2.QtCore import QThread, Signal, QSize, Qt

from output_dialog import show_output

import log
import sys
import io

__all__=["Page","PageCanvas"]

class RedirectIO():
    def __init__(self, s: str = ""):
        self.content = s
    def write(self, s):
        self.content+=str(s)
    def read(self):
        return self.content.strip(" \n")

class RunThread(QThread):
    """
    The real function-executor.
    """
    def __init__(self,func,func_args:tuple=(),func_kwargs:dict={},parent=None, *args, **kwargs):
        super().__init__(parent=None, *args, **kwargs)
        self.func = func
        self.args = func_args
        self.kwargs = func_kwargs
        self.ret=None

    def run(self):
        log.info("run func with args={0},kwargs={1}".format(self.args,self.kwargs))
        self.ret = self.func(*self.args, **self.kwargs)
        return self.ret

class PageCanvas(QWidget):
    OnResize = Signal(QSize)
    def resizeEvent(self, event):
        self.OnResize.emit(self.size())

class Page(QWidget):
    def __init__(self,
            func=None,
            func_args=[],
            func_kwargs={},
            margin=30,
            vert_spacing=10,
            Current_Layout="TopBottomLayout",
            description="",
            *args, **kwargs
        ):
        super().__init__()

        # static definitions
        self.func = func
        self.args = func_args
        self.kwargs = func_kwargs
        self.widget_list = []
        self.margin = margin
        self.vert_spacing =vert_spacing
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        self.Current_Layout=getattr(self,Current_Layout,self.TopBottomLayout)

        # Initiallize: create widgets
        # main_window -> central_widget -> page ->  widget_scroll -> canvas
        self.canvas = PageCanvas()

        # create a scroll bar
        self.widget_scroll = QScrollArea(self)
        self.widget_scroll.setWidget(self.canvas)

        # use a QFormLayout to place widgets
        self.form_layout = QFormLayout()
        self.form_layout.setMargin(self.margin)
        self.form_layout.setSpacing(self.vert_spacing)
        self.form_layout.setRowWrapPolicy(self.form_layout.WrapLongRows)
        self.canvas.setLayout(self.form_layout)

        # default description is func.__doc__
        description = self.func.__doc__ if description=="" else description
        self.description = QLabel(description) # doesn't need to scroll
        self.description.setParent(self)
        self.description.setTextFormat(Qt.MarkdownText)
        self.description.setOpenExternalLinks(True)
        log.info("description geometry={0}".format(self.description.geometry()))

        # create a Run button to excute the function
        self.run_button = QPushButton("&Run", self)
        self.run_button.clicked.connect(self.Run)
        self.run_button.adjustSize()

        self.Current_Layout(self.size())

        log.info("button geometry={geometry}".format(geometry=self.run_button.geometry()))

        # allow the user to run the func multiple times to test output
        # but only the last return is delivered (as a confirmed run)
        self.run_thread = RunThread(self.func, self.args, self.kwargs)
        self.run_thread.finished.connect(self.Done)

    def setArgs(self, *args, **kwargs):
        """
        Set and update arguments for running self.func.
        """
        self.args = args + self.args[len(args):]
        self.kwargs.update(kwargs)

    ### Layouts
    def TopBottomLayout(self, page_size):
        """
        set an up-down layout for self.page.
        """
        self.resize(page_size)

        # description should be placed on the top
        self.description.move(self.x()+self.margin,self.y()+self.margin)
        self.description.adjustSize()

        # run_button should be placed below description
        self.run_button.setGeometry(
            self.margin,
            self.description.y()+self.description.height()+self.margin,
            page_size.width() - 2*self.margin,
            self.run_button.height()
        )

        # canvas should be placed below run_button
        self.canvas.setGeometry(
            self.x(),
            self.run_button.y()+self.run_button.height(),
            page_size.width()-self.margin, # scroll has width
            self.canvas.height()
        )

        # widget_scroll should be placed between the end of page and run_button
        self.widget_scroll.setGeometry(
            self.x(),
            self.run_button.y()+self.run_button.height(),
            page_size.width(),
            page_size.height()-self.run_button.y()-self.run_button.height())

    ### slot functions
    def Run(self):
        """
        Run self.func in another thread.
        """
        for i in self.widget_list:
            self.kwargs.update({i.objectName(): i.getValue()})

        # redirect standard output
        self.redirect_IO = RedirectIO()
        sys.stdout = self.redirect_IO
        sys.stderr = self.redirect_IO

        # avoid multiple clicks
        self.run_button.setText("Running...")
        self.run_button.setEnabled(False)
        self.run_thread.start()

    def Done(self):
        """
        Restore run_button.
        """
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        show_output(self.redirect_IO.read(),self.func.__name__,self.run_thread.ret)
        self.run_button.setText("Run")
        self.run_button.setEnabled(True)


"""
Clrsarisse

Analyze module.
analyze the interface of a function/class.

by 1MLightyears@gmail.com

on 20201208
"""
import inspect
import types_supported
import typing

import log

__all__=["AnalyzeFunc"]

def AddAWidget(widget_type,*args, **kwargs):
    if widget_type is int:
        return types_supported.ClrsInt(*args, **kwargs)
    elif widget_type is float:
        return types_supported.ClrsFloat(*args, **kwargs)
    elif widget_type is str:
        return types_supported.ClrsString(*args, **kwargs)
    elif widget_type is bool:
        return types_supported.ClrsBool(*args,**kwargs)
    elif widget_type is list:
        return types_supported.ClrsList(*args, **kwargs)
    elif widget_type is dict:
        return types_supported.ClrsDict(*args,**kwargs)
    elif isinstance(widget_type,typing._GenericAlias):
        if widget_type._name == "List":
            return types_supported.ClrsList(ElementType=widget_type.__args__, *args, **kwargs)
        elif widget_type._name == "Dict":
            return types_supported.ClrsDict(ElementType=widget_type.__args__, *args, **kwargs)
        else:
            return types_supported.ClrsUnknown(*args, **kwargs)
    else:
        return types_supported.ClrsUnknown(*args, **kwargs)

def AnalyzeFunc(func,*args,**kwargs):
    """
    Analyze the interface of a function, depending on its args and __doc__.
    """
    arguments = inspect.getfullargspec(func)
    log.info("func={0},arguments={1}".format(func.__name__,arguments))
    widget_list=[]
    arg_names=arguments.args or []
    arg_annos = arguments.annotations or []
    arg_defaults=arguments.defaults or []
    for name in arg_names:
        if name in arg_annos:
            widget_list.append(AddAWidget(arg_annos[name], name=name, *args, **kwargs))
            widget_list[-1].adjustSize()
        else:
            widget_list.append(AddAWidget("Unknown", name=name, *args, **kwargs))
            widget_list[-1].adjustSize()
    for i in range(len(arg_defaults)):
        widget_list[-i-1].setDefault(arg_defaults[-i-1])

    # widget_list includes all the widgets needs to be displayed.
    return widget_list



from PySide2.QtWidgets import QDialog, QTextEdit
from PySide2.QtCore import Qt

__all__ = ["show_output"]

class ModifiedQDialog(QDialog):
    def __init__(self, *args, **kwargs):
        super(ModifiedQDialog,self).__init__(*args,**kwargs)
        self.output = QTextEdit(self)
        self.output.setGeometry(0, 0, self.width(), self.height())
        self.output.setReadOnly(True)

    def resizeEvent(self, event):
        super(ModifiedQDialog, self).resizeEvent(event)
        self.output.resize(self.size())

def show_output(s="",func_name="",ret=None):
    dialog = ModifiedQDialog()
    dialog.resize(400,300)
    dialog.output.setText(s)
    dialog.setWindowTitle("Output of {0} - returns {1}".format(func_name,ret))
    dialog.setWindowModality(Qt.NonModal)
    dialog.exec_()
    dialog.deleteLater()


"""
Clarisse

log module.
Process global logging behavior.

by 1MLightyears@gmail.com

on 20201214
"""

import logging
import logging.config
import sys
import json
import os

info = logging.info
warning = logging.warning
error = logging.error
critical = logging.critical
fatal = logging.fatal
debug = logging.debug

### log settings

with open('..'+os.sep+'log_config.json', 'r', encoding='utf-8') as f:
    config = json.load(f)
    logging.config.dictConfig(config)


import setuptools

with open("README.md","r",encoding="utf-8") as fh:
    long_description=fh.read()

setuptools.setup(
        name="clarisse",
        license="Apache 2.0",
        version="0.0.2",
        author="Lightyears",
        author_email="1MLightyears@gmail.com",
        description="A light-weighted GUI framework for Python programs.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url='https://github.com/1MLightyears/clarisse',
        classifiers=[
            "Programming Language :: Python :: 3.9",
            "Operating System :: OS Independent",
            "Development Status :: 2 - Pre-Alpha",
            "License :: OSI Approved :: Apache Software License",
            "Natural Language :: English"
        ],
        install_requires=[
            "PySide2"
        ],
        python_requires='>=3.6',
        packages=setuptools.find_packages(),
)

